# $Id: RCS.tcl,v 3.10 2005-08-23 22:32:34-04 m1tca00 Exp $
#
#   [incr Tcl] Revision Control System interface.
# .SH "SYNOPSIS"
#   *package require RCS*
#   .br
#   *RCS* $objName \fI[CLASS OPTIONS]\fR
#   .br
#   $objName *method* \fI[METHOD OPTIONS]\fR
#
# $Author: m1tca00 $

# $Locker:  $
#
# $Log: RCS.tcl,v $
# Revision 3.10  2005-08-23 22:32:34-04  m1tca00
# More methods are public.
# New method, changeRev, prompts user for new value of -rev.
#
# Revision 3.9  2004/07/19 18:24:42  m1tca00
# More doc cleanup.
#
# Revision 3.8  2004-07-14 00:26:19-04  m1tca00
# Documenation cleanup.
#
# Revision 3.7  2004-07-02 14:32:37-04  m1tca00
# Use blt::bgexec to get old revisions and use $editor to show them.
# WheelMouse bindings for RCSnew and getRevision.
#
# Revision 3.6  2003-09-09 21:08:37-04  m1tca00
# Add WheelMouse bindings for hlist.
#
# Revision 3.5  2003-05-23 14:51:59-04  m1tca00
# Bug fix: if "option get" doesn't return a font, use some default spacing.
#
# Revision 3.4  2003-04-25 10:00:19-04  m1tca00
# Use 'pwd' instead of the solaris-only 'uwd'.
#
# Revision 3.3  2002-12-12 17:41:14-05  m1tca00
# Bug fixes in displaying of symbolic names and locks.
#
# Revision 3.2  2002-12-06 12:09:36-05  m1tca00
# New option -rev (sets symbolic name on checkin).
# getRevision is now a public method (and revReturn/revSelect are public
#   variables).
# _rlog(key,rev) now drops state and lines but includes symbolic names.
# New public method selectFile:
#   Sets active fileName.
#
# Revision 3.1  2001-12-03 09:32:05-05  m1tca00
# Added user-configurable manbrowser.
# Added help method.
# Return helpful error message/version on failure to configure.
#
# Revision 3.0  2001-11-30 17:31:40-05  m1tca00
# Replaced stooop implementation with [incr Tcl].
#

# TODO
#   Allow user to modify -rev

package provide RCS 3.0

# Required extensions
package require Itcl
package require Tclx
package require Tix
package require BWidget
package require menuHelp
package require animCursor
package require BLT
package require GetInput
# Optional... if available, use the images in it
catch {package require imageLib}

itcl::class RCS {
    ####################################################################
    # Configuration options                                            #
    ####################################################################

    # Current directory
    public variable dir  {} {

	# Make sure we can change to this directory
	if {[catch {cd $dir} result]} {
	    error $result
	}
	catch {getRCSinfo}
    }

    # Background refresh rate
    public variable refresh

    # Editor
    public variable editor

    # Man page browser
    public variable manbrowser

    # Symbolic revision name
    public variable rev {} {
	if {[string length $rev]} {
	    set rev -N$rev
	}
    }

    ####################################################################
    # Private variables                                                #
    ####################################################################

    # Version (set by RCS)
    private variable version {$Revision: 3.10 $}

    # GUI components
    private variable top
    private variable buttons
    private variable status
    private variable menu
    private variable text
    private variable hlist

    # Tix styles
    private variable head_style
    private variable h_style0
    private variable h_style1

    #
    # State variables
    #

    # Status message
    private variable statusMsg
    # Scan context
    private variable context
    # Descriptive text
    private variable descReturn
    private variable descText
    # Selected revision(s)
    public variable revReturn
    public variable revSelect
    # Rlog array
    private variable _rlog
    # bgexec status/stdout/stderr
    private variable bg_status {}
    private variable bg_output {}
    private variable bg_error {}
    # Temporary directory name
    private variable _tmpDir {}

    ####################################################################
    # Constructor/Destructor                                           #
    ####################################################################

    constructor {args} {}
    destructor {}

    # RCS::version
    #
    #	    Return the version number.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The RCS version number is returned.
    public method version {}

    # RCS::help
    #
    #	    Return the help text.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Help intro is returned.
    public method help {}

    ####################################################################
    # User preferences                                                 #
    ####################################################################

    # RCS::preferences
    #
    #	    Initialize preferences.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Default resources are set and user resources are read from
    #	    "~/.tkrcs-defaults".  Nothing will be returned.
    private method preferences {}

    # RCS::fileMenu
    #
    #	    Build the File menu
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The $menu will be built.  Nothing will be returned.
    private method fileMenu {}

    # RCS::backgroundUpdate
    #
    #	    Update the $hlist and schedule next execution.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Update the $hlist and reschedule execution after $refresh.
    #	    Nothing will be returned.
    private method backgroundUpdate {}

    # RCS::contextInit
    #
    #	    Initialize a scan context for parsing rlog.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The scan context will be created.  Nothing will be returned.
    private method contextInit {}

    # RCS::getRCSinfo
    #
    #	    Get a list of RCS files in specified directory.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Scan the RCS files in the current directory and build the
    #	    _rlog array.  Return 0 on success or 1 on error.
    public method getRCSinfo {}

    # RCS::updateHlist
    #
    #	    Update the listbox.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The $hlist will be updated with the current $_rlog array.
    #	    Nothing will be returned.
    private method updateHlist {}

    # RCS::inOrOut
    #
    #	    Check key back in or out (depending on current state).
    #
    # Arguments:
    #	    key		File to check in/out (optional, use current
    #			seleection if null).
    #
    # Results:
    #	    If $key is locked, check the file in, otherwise check it
    #	    out.  If $key is null, check to see if a current selection
    #	    exists and use it.	Once run, execute getRCSinfo.  Return 0
    #	    on success or 1 on error.
    public method inOrOut {{key {}}}

    # RCS::setMenu
    #
    #	    Enable/disable menu entries based on current selection.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    If nothing is selected, disable file-based menu entries.
    #	    Otherwise, enable them and set the "check in/out" label
    #	    based on lock state of current selection.  Nothing will be
    #	    returned.
    private method setMenu {}

    # RCS::showInfo
    #
    #	    Show info from rlog in text window.
    #
    # Arguments:
    #	    key		File to show info about (optional).
    #
    # Results:
    #	    If $key is not null, show information in the _rlog array
    #	    about it, otherwise do nothing.  Nothing will be returned.
    public method showInfo {{key {}}}

    # RCS::showText
    #
    #	    Show data in the text widget.
    #
    # Arguments:
    #	    textStr	String to show.
    #	    wrap	Wrap mode for text widget (optional, default
    #			"none").
    #
    # Results:
    #	    $textStr will be displayed in the $text widget and
    #	    configured to wrapmode $wrap.  Nothing will be returned.
    public method showText {textStr {wrap none}}

    # RCS::getCurrent
    #
    #	    Return the current selection in hlist (if any).
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The current key will be returned (or null if there is no
    #	    current key).
    public method getCurrent {}

    ####################################################################
    # Interface to RCS commands                                        #
    ####################################################################

    # RCS::RCSident
    #
    #	    Identify keywords in the file.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    ident will be executed on the current selection and results
    #	    will be passed to showText.  Nothing will be returned.
    public method RCSident {}

    # RCS::RCSrlog
    #
    #	    Show the log of all changes to the file.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    rlog will be executed on the current selection and results
    #	    will be passed to showText.  Nothing will be returned.
    public method RCSrlog {}

    # RCS::RCSview
    #
    #	    Show the file and allow user to see previous versions.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    rcsview will be executed on the current selection.	Nothing
    #	    will be returned.
    public method RCSview {}

    # RCS::TKdiff
    #
    #	    Compare the current file with the latest revision.
    #
    # Arguments:
    #	    selectVers	    Allow user to select versions (boolean,
    #			    optional, default false).
    #
    # Results:
    #	    If $selectVers is true, user will be prompted to select two
    #	    versions.  Otherwise, the current file will be compared to
    #	    the head revsision.  tkdiff will be executed to compare the
    #	    two revisions.  Return 0 on success or 1 on error.
    public method TKdiff {{selectVers 0}}

    # RCS::RCSedit
    #
    #	    Edit the file.
    #
    # Arguments:
    #	    key		File to be edited (optional, default is current
    #			selection).
    #
    # Results:
    #	    If $key is not null, it will be set to the current
    #	    selection.	The $editor will be executed on the file.
    #	    Nothing will be returned.
    public method RCSedit {{key {}}}

    # RCS::RCSshowComments
    #
    #	    Show the comments from the top of the file.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Using ::menuHelp::readHelpFile, the comments at the head of
    #	    the current file (if any) will be displayed in the $text
    #	    widget.  Nothing will be returned.
    public method RCSshowComments {}

    # RCS::RCSmodifyLog
    #
    #	    Allow the user to modify log comments from an earlier
    #	    revision.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    User will be prompted for an old revision.	Once selected,
    #	    user will be able to edit the log message from that revision
    #	    and commit the change.  Return 0 on success or 1 if no
    #	    changes are made.
    public method RCSmodifyLog {}

    # RCS::RCSviewOld
    #
    #	    View a previous revision.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    User will be prompted for an old revision.  Once selected,
    #	    that revision will be show to the user.  Return 0 on success
    #	    or 1 if failure.
    public method RCSviewOld {}

    # RCS::RCScheckinNew
    #
    #	    Check in and specify revision.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    User will be prompted to select the revision and for a log
    #	    message.  Once selected, the new revision will be checked
    #	    in.  Return 0 on success or 1 if no changes are made.
    public method RCScheckinNew {}

    # RCS::RCSoutDate
    #
    #	    Delete (outdate) older revisions.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    User will be prompted for (multiple, continuos) revisions.
    #	    All revisions in the selected range will be "outdated" from
    #	    RCS.  Return 0 on success or 1 if no changes are made.
    public method RCSoutDate {}

    # RCS::RCSsetBranch
    #
    #	    Set the default branch.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    User will be prompted for an old revision.  That revision
    #	    will be made the default branch in RCS.  Return 0 on
    #	    success or 1 if no changes are made.
    public method RCSsetBranch {}

    ####################################################################
    # DIALOGS                                                          #
    ####################################################################

    # RCS::RCSnew
    #
    #	    Check in a new file.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    If no RCS directory exists, prompt user if one should be
    #	    created.  User will be prompted to select a file in $dir
    #	    that is not already checked in.  Return 0 on success or 1 if
    #	    no changes are made.
    public method RCSnew {}

    # RCS::RCScreate
    #
    #	    Check in a new file.
    #
    # Arguments:
    #	    args	List of files to check in.
    #
    # Results:
    #	    Called by RCSnew.  Loop through $args and display each file
    #	    and exec ci on the file.  Return 0 on success or 1 on
    #	    failure.
    public method RCScreate {args}

    # RCS::changeDir
    #
    #	    Prompt user and change current directory.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Nothing will be returned.
    public method changeDir {}

    # RCS::changeRev
    #
    #	    Prompt user and change symbolic name.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    Nothing will be returned.
    public method changeRev {}

    # RCS::setCBox
    #
    #	    Validate change to the combo box.
    #
    # Arguments:
    #	    d		New directory (optional, default null).
    #
    # Results:
    #	    The combobox widget will attempt to set the value to
    #	    $d.  The $_tmpDir variable will be set to the value of
    #	    the current directory ($d if validation was successful).
    #	    Nothing will be returned.
    private method setCBox {{d {}}}

    # RCS::dirOK
    #
    #	    Called by the OK button in changeDir.
    #
    # Arguments:
    #	    none.
    #
    # Results:
    #	    The selected directory is validated and configured.  The
    #	    .tkrcs-dirs list will be updated to include the new
    #	    directory.  The animated cursor will stop, the dialog will
    #	    be withdrawn, and focus will be released.  Return 0 on
    #	    success or 1 on failure.
    private method dirOK {}

    # RCS::getDescription
    #
    #
    #	    Allow user to enter or edit a description (change log).
    #
    # Arguments:
    #	    prompt	Prompt to show user.
    #	    curText	Current text to allow editing (optional, default
    #			is null).
    #
    # Results:
    #	    User will be prompted to enter/edit text.  New text will be
    #	    stored in $descText and the return code will be stored in
    #	    $descReturn.  Return 0 if text is ok or 1 if not (cancelled,
    #	    etc).
    public method getDescription {prompt {curText {}}}

    # RCS::getDescText
    #
    #	    Filter text from the widget and make required quoting
    #	    substitions.
    #
    # Arguments:
    #	    t		Name of text widget.
    #
    # Results:
    #	    The formatted text from the $t text widget is returned.
    private method getDescText {t}

    # RCS::getRevision
    #
    #	    Allow user to select a revision or enter a new one.
    #
    # Arguments:
    #	    newRev	Allow user to enter a revision (boolean,
    #			optional, default false).
    #	    browseMode	Mode to set listbox into (optional, default
    #			browse).
    #
    # Results:
    #	    The user will be allowed to select one or more old revisions
    #	    (depending on $browseMode) or enter a new revision
    #	    (depending on $newRev).  The selected revision(s) will be
    #	    stored in $revSelect and the return code will be stored in
    #	    revReturn.  Return 0 if ok or 1 if not (cancelled, etc).
    public method getRevision {{newRev 0} {browseMode browse}}

    # RCS::selectFile
    #
    #	    Allow a user to select a file in the current directory.
    #
    # Arguments:
    #
    #	    fileName	Name of the file to select.
    #
    # Results:
    #	    The $fileName will be set in the $hlist.  Nothing will be
    #	    returned.  Errors will propogate.
    public method selectFile {fileName}
}

####################################################################
#                                                                  #
# Constructor/Destructor                                           #
#                                                                  #
####################################################################

itcl::body RCS::constructor {args} {
    # Create a toplevel window
    set top .rcs_[namespace tail $this]
    destroy $top
    toplevel $top -relief ridge -class Rcs
    wm title $top "Revision Control System"

    # Make sure to delete THIS object if the window is destroyed
    wm protocol $top WM_DELETE_WINDOW "catch {itcl::delete object $this}"
    bind $top <Destroy> "catch {itcl::delete object $this}"

    # Get preferences from resources
    preferences

    # Set default directory and configure user options
    set dir [pwd]
    if {[catch {eval configure -dir $dir $args}]} {
	catch {wm withdraw $top}
	if {[lcontain $args "--version"]} {
	    error "Version [version]\n\nWritten by Tom Allard <tallard@frb.gov>"
	} else {
	    error "[help]\n\nOptions:\n [join [configure] "\n "]"
	}
    }

    # Buttons frame
    set buttons [frame $top.buttons -relief raised -borderwidth 4]
    menubutton $buttons.file -text File -relief raised -takefocus 1 \
	-highlightthickness 2 -highlightcolor blue \
	-menu $buttons.file.menu
    button $buttons.exit -text Exit -relief raised -takefocus 1 \
	-highlightthickness 2 -highlightcolor blue \
	-command "itcl::delete object $this"

    # Try and configure images created by imageLib
    catch "imageLib::imageConfig $buttons.exit Exit -relief flat"
    catch "imageLib::imageConfig $buttons.file File -relief flat"

    # Status label
    set status [label $buttons.status \
	-textvariable [itcl::scope statusMsg] \
	-foreground darkgreen -padx 3 -justify left]

    # File menu
    set menu [menu $buttons.file.menu]
    bind $top <Alt-f> "
	tkMbPost $buttons.file
	tkMenuFirstEntry $menu
    "

    #
    # Created Paned Window (for text and hlist)
    #
    tixPanedWindow $top.pane -paneborderwidth 4 -panerelief groove \
	-separatorbg red -handleactivebg red \
	-orientation vertical
    # Size depends on font...
    if {[string length [option get $top font Font]]} {
	set spacing [font metrics [option get $top font Font] -linespace]
    } else {
	set spacing 14
    }
    $top.pane add text \
	-min [expr 8*$spacing] \
	-size [expr round(13.5*$spacing)]
    $top.pane add hlist \
	-min [expr 6*$spacing] \
	-size [expr 23*$spacing]
    set ptext [$top.pane subwidget text]
    set phlist [$top.pane subwidget hlist]

    # Scrolled Text for documentation, log, etc
    tixScrolledText $ptext.text -scrollbar both
    set text [$ptext.text subwidget text]
    $text configure -width 40 -height 10 -relief groove \
	-takefocus 0 -state disabled -wrap none -foreground blue \
	-cursor {}

    # Scrolled Hlist of RCS files/status
    tixScrolledHList $phlist.rcs \
	-options "
	    hlist.command	{[itcl::code $this inOrOut]}
	    hlist.browseCmd	{[itcl::code $this showInfo]}
	    hlist.columns	2
	    hlist.header	1
	    hlist.selectMode	browse
	    hlist.width		60
	    hlist.height	20
	    hlist.separator	/
	    hlist.indent	20
	"
    $phlist.rcs configure -highlightcolor blue
    set hlist [$phlist.rcs subwidget hlist]

    # Create styles
    set head_style [tixDisplayStyle text -refwindow $hlist -fg black]
    set h_style0 \
	[tixDisplayStyle text -refwindow $hlist \
	    -foreground darkgreen -selectforeground darkgreen]
    set h_style1 \
	[tixDisplayStyle text -refwindow $hlist \
	    -foreground red -selectforeground red]

    # Configure Headings
    $hlist header create 0 -itemtype text \
	-text "Checked In" -style $head_style
    $hlist header create 1 -itemtype text \
	-text "Checked Out/Locked" -style $head_style
    $hlist header configure 0 -relief groove
    $hlist header configure 1 -relief groove
    $hlist column width 1 0
    $hlist column width 0 -char 30
    $hlist column width 1 -char 30

    # Bindings
    bind $top <Control-q> "itcl::delete object $this"
    bind $hlist <Key-Home> "$hlist yview moveto 0"
    bind $hlist <Key-End> "$hlist yview moveto 1"

    # Wheel mouse
    bind $hlist <Button-4> "%W yview scroll -5 units ; break"
    bind $hlist <Button-5> "%W yview scroll 5 units ; break"
    bind $hlist <Shift-Button-4> "%W yview scroll -1 units ; break"
    bind $hlist <Shift-Button-5> "%W yview scroll 1 units ; break"
    bind $hlist <Control-Button-4> "%W yview scroll -1 pages ; break"
    bind $hlist <Control-Button-5> "%W yview scroll 1 pages ; break"
    bind $hlist <Alt-Button-4> "%W xview scroll -5 units ; break"
    bind $hlist <Alt-Button-5> "%W xview scroll 5 units ; break"

    # Pack windows
    pack $buttons -side top -fill both -expand 0 -pady 4
    pack $buttons.exit -side right
    pack $buttons.file -side left
    pack $status -side left
    pack $top.pane -fill both -expand 1
    pack $ptext.text -fill both -expand 1
    pack $phlist.rcs -fill both -expand 1

    # Build the file menu
    fileMenu

    # Initialize scancontext for rlog
    contextInit

    # Load RCS info
    backgroundUpdate

    # Show some initial help
    showText [help] word
}

itcl::body RCS::destructor {} {
    # Clear after jobs
    set myAfter "\{$this"
    foreach a [after info] {
	if {[regexp -- $myAfter [after info $a]]} {
	    after cancel $a
	}
    }

    # Delete any menuHelp (if any)
    catch {itcl::delete object ${this}_help}

    # Reset animCursor
    animCursor::resetCursor

    # Delete old context
    catch "scancontext delete $context"

    # Unbind and destroy $top (if any)
    catch "bind $top <Destroy> {}"
    catch "destroy $top"
}

# Return the version number
itcl::body RCS::version {} {
    return [lindex $version 1]
}

# Return the help text
itcl::body RCS::help {} {
    return {The Revision Control System (RCS) manages multiple\
revisions of files.  RCS automates the storing,\
retrieval, logging, identification, and merging of\
revisions.

This is a front-end to the most common RCS functions.

Select "File/Help" for assistance or
"File/Man Pages" for info on RCS.}
}

########################################################################
#                                                                      #
# User preferences                                                     #
#                                                                      #
########################################################################

# Initialize preferences.
itcl::body RCS::preferences {} {
    # Automatically reread RCS every minute
    option add *Rcs.refresh	    60000			20
    # Editor
    option add *Rcs.editor	    /bks/proj/bin/nedit		20
    # Manual browser
    option add *Rcs.manbrowser	    /bks/proj/bin/tkman.remote	20
    # Default font for check in descriptions
    option add *Rcs.desc.text*Font  -*-fixed-bold-*-*-*-*-*-*-*-*-*-*-* 85
    # A few color resources I don't like from Tix
    option add *Rcs*activeBackground gray95			interactive
    option add *Rcs*activeForeground darkgreen			interactive
    option add *Rcs*selectBackground lightblue			interactive
    option add *Rcs*selectForeground blue			interactive
    option add *Rcs*Menu*background  lightblue			interactive
    option add *Rcs*Menu*foreground  blue			interactive
    option add *Rcs*troughColor      pink			interactive

    # If the user has any preferences, they can put them here...
    catch "option readfile ~/.tkrcs-defaults 90"

    # Set the refresh rate and return
    foreach opt {refresh editor manbrowser} {
	set [set opt] [option get $top $opt {}]
    }
    return
}

# Build the File menu
itcl::body RCS::fileMenu {} {
    # Common widget names
    set m $menu

    #
    # Populate the File menu
    #

    # First block acts on current selection
    $m add command -label "Check In/Out" -underline 0 -command "
	#
	#Check the file in or out (depending on current state).
	#
	#Check in with:
	#  ci -zLT -u FILENAME
	#
	#Check out with:
	#  co -zLT -l FILENAME
	[itcl::code $this inOrOut]
    "
    bind $top <Key-c> "[itcl::code $this inOrOut]"

    $m add command -label "Identify" -underline 0 -command "
	#
	#Identify keywords in the file.
	#
	#  ident FILENAME
	[itcl::code $this RCSident]
    "
    bind $top <Key-i> "[itcl::code $this RCSident]"

    $m add command -label "Log" -underline 0 -command "
	#
	#See all log messages and details of the RCS-file.
	#
	#  rlog -zLT FILENAME
	[itcl::code $this RCSrlog]
    "
    bind $top <Key-l> "[itcl::code $this RCSrlog]"

    $m add command -label "rcsview" -underline 3 -command "
	#
	#rcsview allows you to easily
	#view the different versions
	#of the file.
	#
	#  rcsview FILENAME
	[itcl::code $this RCSview]
    "
    bind $top <Key-v> "[itcl::code $this RCSview]"

    $m add command -label "tkdiff" -underline 0 -command "
	#
	#See side-by-side differences of working file and last check-in.
	#
	#  tkdiff -r FILENAME
	[itcl::code $this TKdiff]
    "
    bind $top <Key-t> "[itcl::code $this TKdiff]"

    $m add command -label "Edit" -underline 0 -command "
	#
	#Edit the file.
	#
	#  [option get $top editor {}] FILENAME
	#
	#You can change the editor by adding to your ~/.tkrcs-defaults:
	#
	# *Rcs.editor: \[editor command\]
	[itcl::code $this RCSedit]
    "
    bind $top <Key-e> "[itcl::code $this RCSedit]"
    # Bind ButtonRelease-3 to edit as well...
    #   Uses menuHelp::bindButton internal to get the key.
    bind $hlist <ButtonRelease-3> "
	_tixRecordFlags <ButtonRelease-3>  %% %# %b %c %d %f %h %k %m %o %p \
		%s %t %w %x %y %A %B %E %K %N %R %T %W %X %Y
	[itcl::code $this RCSedit] \[menuHelp::bindButton %W %x %y\]
	_tixDeleteFlags
    "

    $m add command -label "Show Leading Comments" -underline 0 -command "
	#
	#Show the comments at the head of the file.
	[itcl::code $this RCSshowComments]
    "
    bind $top <Key-s> "[itcl::code $this RCSshowComments]"

    # MORE RCS commands...
    $m add cascade -label "More..." -underline 0 -menu $m.more
    menu $m.more
    $m.more add command -label "Replace old log message" -command "
	#
	#Modify the log message of a previous revision.
	#
	#  rcs -mREV:MESSAGE FILENAME
	[itcl::code $this RCSmodifyLog]
    "

    $m.more add command -label "View old revision" -command "
	#
	#View an earlier revision.
	#
	#  co -pREV FILENAME
	[itcl::code $this RCSviewOld]
    "

    $m.more add command -label "Check in and specify revision" -command "
	#
	#Check in file and specify revision.
	#
	#  ci -zLT -f -uREV FILENAME
	[itcl::code $this RCScheckinNew]
    "

    $m.more add command -label "Delete (outdate) old revisions" -command "
	#
	#Remove old revisions from RCS.
	#
	#  rcs -oREV FILENAME
	[itcl::code $this RCSoutDate]
    "

    $m.more add command -label "Set default branch" -command "
	#
	#Set the default branch.
	#
	#  rcs -bREV FILENAME
	[itcl::code $this RCSsetBranch]
    "

    $m.more add command -label "Compare any two revisions" -command "
	#
	#Use tkdiff to compare any two selected revisions.
	#
	#  tkdiff -rREV1 -rREV2 FILENAME
	[itcl::code $this TKdiff] 1
    "

    # This block does NOT require current selection
    $m add separator
    $m add command -label "Check in New File" -underline 9 -command "
	#
	#Add a file in the current directory to the RCS system.
	#
	#  ci -zLT -u FILENAME
	#
	#  (If the RCS subdirectory
	#  does not already exist,
	#  you will be asked if one
	#  should be created.)
	[itcl::code $this RCSnew]
    "
    bind $top <Key-n> "[itcl::code $this RCSnew]"

    $m add command -label "Change Directory" -underline 7 -command "
	#
	#Allow you to select a new directory.
	[itcl::code $this changeDir]
    "
    bind $top <Key-d> "[itcl::code $this changeDir]"

    $m add command -label "Refresh List" -underline 0 -command "
	#
	#Force the system to check for locked/unlocked files.
	#
	#This is automatically called once a minute.
	[itcl::code $this getRCSinfo]
    "
    bind $top <Key-r> "[itcl::code $this getRCSinfo]"

    $m add command -label "Change Symbolic Name" -underline 2 -command "
	#
	#Change the symbolic name.
	#
	#The symoblic name is initially set with the -rev option.
	[itcl::code $this changeRev]
    "
    bind $top <Key-a> "[itcl::code $this changeRev]"

    $m add separator
    $m add command -label "Help" -underline 0 -command "
	#
	#Browse the File menu and see Help text (this window).
	#
	#Selecting an entry will show help for that command.
	#
	#Double-click or <Return> to execute command.
	catch {itcl::delete object ${this}_help}
	menuHelp ${this}_help \
	    -title {RCS File} \
	    -menu $m \
	    -helpLoaded 1
    "
    bind $top <Key-h> "
	catch {itcl::delete object ${this}_help}
	menuHelp ${this}_help \
	    -title {RCS File} \
	    -menu $m \
	    -helpLoaded 1
    "

    # Manul pages on RCS
    $m add cascade -label "Man Pages" -underline 4 -menu $m.man
    menu $m.man
    $m.man add command -label "rcsintro(1)" -command "
	#
	#introduction to RCS commands
	system {$manbrowser rcsintro.1} &
    "

    $m.man add command -label "ci(1)" -command "
	#
	#check in RCS revisions
	system {$manbrowser ci.1} &
    "

    $m.man add command -label "co(1)" -command "
	#
	#check out RCS revisions
	system {$manbrowser co.1} &
    "

    $m.man add command -label "ident(1)" -command "
	#
	#identify RCS keyword strings in files
	system {$manbrowser ident.1} &
    "

    $m.man add command -label "rlog(1)" -command "
	#
	#print log messages and other information about RCS files
	system {$manbrowser rlog.1} &
    "

    $m.man add command -label "rcs(1)" -command "
	#
	#change RCS file attributes
	system {$manbrowser rcs.1} &
    "

    $m.man add command -label "rcsdiff(1)" -command "
	#
	#compare RCS revisions
	system {$manbrowser rcsdiff.1} &
    "

    $m.man add command -label "rcsmerge(1)" -command "
	#
	#merge RCS revisions
	system {$manbrowser rcsmerge.1} &
    "

    $m.man add command -label "rcsfile(5)" -command "
	#
	#format of RCS file
	system {$manbrowser rcsfile.5} &
    "

    # End of menu
    $m add separator
    $m add command -label "Exit" -underline 1 -command "
	#
	#Exit
	itcl::delete object $this
    "
}

# Update the $hlist and schedule next execution.
itcl::body RCS::backgroundUpdate {} {
    # Clear after jobs
    set myAfter "\{$this"
    foreach a [after info] {
	if {[regexp -- $myAfter [after info $a]]} {
	    after cancel $a
	}
    }
    # Update RCS info and reschedule
    getRCSinfo
    after $refresh "catch {[itcl::code $this backgroundUpdate]}"
    return
}

# Initialize a scan context for parsing rlog.
itcl::body RCS::contextInit {} {
    set context [scancontext create]

    # New RCS file/head
    scanmatch $context {^RCS file:} {
	set f [lindex $matchInfo(line) end]
	regsub -- {,v$} [file tail $f] {} key
	set _rlog($f,key) $key
	set _rlog($key,RCSfile) $f
	set _rlog($key,locked) 0
    }
    scanmatch $context {^head:} {
	set _rlog($key,version) [lindex $matchInfo(line) end]
	set _rlog($key,revHist) {}
    }

    # Beginning of symbolic names
    scanmatch $context {^symbolic names:} {
	set _rlog(symnames) 1
	foreach k [lmatch -regexp [array names _rlog] "^$key,sym"] {
	    set _rlog($k) {}
	}
    }

    # Symbolic names
    scanmatch $context {^\t} {
	if {[info exists _rlog(symnames)]} {
	    lappend _rlog($key,sym,[lindex $matchInfo(line) 1]) \
		[lindex [split $matchInfo(line) \t:] 1]
	}
    }

    # End of symnames names
    scanmatch $context {^keyword substitution:} {
	if {[info exists _rlog(symnames)]} {
	    unset _rlog(symnames)
	}
    }

    # Revision/date
    scanmatch $context {^revision } {
	set lastRev [lindex $matchInfo(line) 1]
	if {[string equal "locked" [lindex $matchInfo(line) 2]]} {
	    set _rlog($key,locked) 1
	    set _rlog($key,$lastRev,lockedBy) [lrange $matchInfo(line) 2 end]
	} else {
	    set _rlog($key,$lastRev,lockedBy) {}
	}
	lappend _rlog($key,revHist) $lastRev
	set _rlog($key,$lastRev) {}
    }

    scanmatch $context {^date:} {
	foreach el [split $matchInfo(line) {\;}] {
	    if {[regexp {(date:|author:)} $el]} {
		append _rlog($key,$lastRev) "$el;"
	    }
	}
	if {[info exists _rlog($key,sym,$lastRev)]} {
	    foreach sym $_rlog($key,sym,$lastRev) {
		append _rlog($key,$lastRev) "  symbolic name: $sym;"
	    }
	}
	if {[string length $_rlog($key,$lastRev,lockedBy)]} {
	    append _rlog($key,$lastRev) "\n\t$_rlog($key,$lastRev,lockedBy)"
	}
    }

    return
}

# Get a list of RCS files in specified directory.
itcl::body RCS::getRCSinfo {} {
    # Find old keys
    set OldKeys [lmatch -regexp [array names _rlog] "^.*,locked\$"]
    # Remove old keys
    foreach Child $OldKeys {
	# Get the base name and add to the listbox
	set key "[lindex [split $Child ","] 0]"
	set DeleteMe [lmatch -regexp [array names _rlog] "^$key,.*\$"]
	foreach k $DeleteMe {
	    unset _rlog($k)
	}
    }

    # Check for RCS subdirectory.
    # Return 1 if no RCS directory exsits.
    if {![file isdirectory $dir/RCS]} {
	set statusMsg "$dir/RCS does not exist!"
	$status configure -foreground red
	updateHlist
	return 1
    }

    # Update status message
    set statusMsg "$dir"
    $status configure -foreground darkgreen

    # Store rlog output here
    set rlogFile /tmp/rlog_$top.[pid]

    # Find all RCS-files
    set RCSfiles [lsort \
	    [glob -nocomplain -- $dir/RCS/* $dir/RCS/.\[a-zA-Z0-9_\]*]]
    if {[system "rlog -zLT $RCSfiles > $rlogFile 2>&1"]} {
	showText [read_file -nonewline $rlogFile]
	set statusMsg "Error:"
	$status configure -foreground red
	set Files {}
    } else {
	set fid [open $rlogFile]
	scanfile $context $fid
	close $fid
    }
    file delete -force $rlogFile

    # Update the listbox
    updateHlist

    # Return a 0 (no errors)
    return 0
}

# Update the listbox.
itcl::body RCS::updateHlist {} {
    # Try and remember the current selection...
    set CurSelection [$hlist selection get]

    # Clear the listbox
    foreach Child [$hlist info children] {
	$hlist delete entry $Child
    }

    # Find matches in _rlog array
    set RCSfiles [lmatch -regexp [array names _rlog] "^.*,locked\$"]
    # Add matches to listbox
    foreach Child [lsort $RCSfiles] {
	# Get the base name and add to the listbox
	set key "[lindex [split $Child ","] 0]"
	$hlist add $key -itemtype text -text {}
	# Add to apropriate column (Checked In/Out)
	$hlist item create $key $_rlog($key,locked) \
	    -itemtype text -text $key \
	    -style [set h_style$_rlog($key,locked)]
    }

    # If CurSelection is still available, we want to select it again
    if {[$hlist info exists $CurSelection]} {
	$hlist selection clear
	$hlist selection set $CurSelection
	$hlist anchor set $CurSelection
    }

    # Set menu depending on selection
    setMenu
    return
}

# Check key back in or out (depending on current state).
itcl::body RCS::inOrOut {{key {}}} {
    # Check to see if key was passed...
    if {![string length $key]} {
	# No key, try the current key (if any)
	set key [getCurrent]
	if {![string length $key]} {
	    # No key, can't procede
	    return 1
	}
    }
    if {[catch "cd $dir"]} {
	set statusMsg "$dir/RCS does not exist!"
	$status configure -foreground red
	return 1
    }

    # Check in or check out...
    if {$_rlog($key,locked)} {
	# File is locked, so we check it in...
	showText "Checking $key in..."
	# If there are differences, run tkdiff first...
	catch "exec rcsdiff -zLT -q $key" result
	if {[string length $result]} {
	    # View the differences
	    system "tkdiff -r $key" &
	    after 5000
	    # Prompt user for log
	    if {[getDescription "Enter log message"]} {
		# Canceled, exit
		showText "Check in canceled."
		return 1
	    } else {
		set logText $descText
	    }
	} else {
	    showText "\nNo differences..."
	    set logText {no message}
	}
	update idletasks
	catch "exec ci -zLT $rev \"-m$logText\" -u $key < /dev/null" result
	showText "$logText\n\n$result"
    } else {
	# File is unlocked, so we check it out...
	catch "exec co -zLT -l $key < /dev/null" result
	showText "$result"
    }

    # Refresh RCS info
    getRCSinfo
    return 0
}

# Enable/disable menu entries based on current selection.
itcl::body RCS::setMenu {} {
    # Common widget names
    set m $menu

    # Determine current selection
    set key [$hlist selection get]

    if {![string length $key]} {
	# No key, so disable menu entries from 1 to seperator and return
	set j 1
	while {"[$m type $j]" != "separator"} {
	    $m entryconfigure $j -state disabled
	    incr j
	}
	$m entryconfigure 1 -label "Check In/Out"
	return
    } else {
	# Enable menu entries
	set j 1
	while {"[$m type $j]" != "separator"} {
	    $m entryconfigure $j -state normal
	    incr j
	}
    }

    # Set menu entry 1 label based on current lock state
    if {$_rlog($key,locked)} {
	# File is locked, so allow user to check IN
	$m entryconfigure 1 -label "Check In (ci -u)"
    } else {
	# File is NOT locked, so allow user to check OUT
	$m entryconfigure 1 -label "Check Out (co -l)"
    }
    return
}

# Show info from rlog in text window.
itcl::body RCS::showInfo {{key {}}} {
    # First, set the menu
    setMenu

    # If no key, do nothing and return
    if {![string length $key]} {
	return
    }

    # Set info to display
    set info "\n$_rlog($key,RCSfile) $_rlog($key,version)\n"
    append info "\nRevision history:\n"
    foreach r $_rlog($key,revHist) {
	append info " [lindex $r 0]\t$_rlog($key,$r)\n"
	if {[llength $r] > 1} {
	    append info "\t[lrange $r 1 end]\n"
	}
    }
    showText "$info"
    return
}

# Show data in the text widget.
itcl::body RCS::showText {textStr {wrap none}} {
    # Delete old text and insert the new
    $text configure -state normal
    $text delete 0.0 end
    $text insert insert $textStr
    $text configure -state disabled -wrap $wrap
    return
}

# Return the current selection in hlist (if any).
itcl::body RCS::getCurrent {} {
    # Get RCS file
    if {[catch "cd $dir"]} {
	showText "Cannot CD to $dir"
	return
    }
    set key [$hlist selection get]
    if {![string length $key]} {
	# If no file, print warning and return null
	showText "No file selected..."
	return
    }
    return $key
}

########################################################################
#                                                                      #
# Interface to RCS commands                                            #
#                                                                      #
########################################################################

# Identify keywords in the file.
itcl::body RCS::RCSident {} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return
    }

    # Run ident and print results
    catch "exec ident $key" result
    showText "$result"
    return
}

# Show the log of all changes to the file.
itcl::body RCS::RCSrlog {} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return
    }

    # Run rlog and print results
    catch "exec rlog -zLT $key" result
    showText "$result"
    return
}

# Show the file and allow user to see previous versions.
itcl::body RCS::RCSview {} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return
    }

    # Run rcsview
    showText "Running rcsview..."
    system "rcsview $key" &
    return
}

# Compare the current file with the latest revision.
itcl::body RCS::TKdiff {{selectVers 0}} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }

    # Does user want to select versions?
    if {$selectVers} {
	showText \
	    "\nSelect two revisions.\n(Ctrl-Button-1 to add the second.)"
	# Get revisions to compare
	if {[getRevision 0 extended]} {
	    showText "tkdiff cancelled."
	    return 1
	}
	# Make sure exactly two were selected
	if {[llength $revSelect] != 2} {
	    showText "tkdiff can only compare two revisions."
	    return 1
	}
	# Run tkdiff
	set r1 [lindex $revSelect 1]
	set r2 [lindex $revSelect 0]
	showText "Running tkdiff..."
	system "tkdiff -r$r1 -r$r2 $key" &
	return 0
    } else {
	# Run tkdiff (default will compare current file with head rev)
	showText "Running tkdiff..."
	system "tkdiff -r $key" &
	return 0
    }
}

# Edit the file.
itcl::body RCS::RCSedit {{key {}}} {
    # Try and find the edit file
    if {[string length $key]} {
	$hlist selection clear
	$hlist selection set $key
	$hlist anchor set $key
    }

    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return
    }

    # Run editor
    showText "Editing file..."
    system "$editor $key" &
    return
}

# Show the comments from the top of the file.
itcl::body RCS::RCSshowComments {} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return
    }

    # Read comments and display
    showText "[menuHelp::readHelpFile $dir/$key]" word
    return
}

# Allow the user to modify log comments from an earlier revision.
itcl::body RCS::RCSmodifyLog {} {
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }
    showText "Modifying log..."

    # Ask user which selection they want
    if {[getRevision]} {
	showText "Modify log cancelled."
	return 1
    }
    set r $revSelect

    # Get current log entry, if possible
    if {[catch "exec rlog -zLT -r$r $key | sed -e 1,/^date:/d" result]} {
	set result {}
    } else {
	regsub -all "\n=+" $result "\n" result
    }

    # Show the prompt
    if {[getDescription "New log for revision $r:" $result]} {
	showText "Modify log cancelled."
	return 1
    } else {
	set logText $descText
    }
    catch "exec rcs \"-m$r:$logText\" $key < /dev/null" result
    showText $result
    return 0
}

# View a previous revision.
itcl::body RCS::RCSviewOld {} {
    showText "Viewing old revision..."
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }

    # Ask user which selection they want
    if {[getRevision]} {
	showText "View old cancelled."
	return 1
    }
    set r $revSelect

    # Check out old revision and run xless
    if {[catch {blt::bgexec $bg_status \
		-error [itcl::scope bg_error] \
		-output [itcl::scope bg_output] \
		co -zLT -p$r $key} result]} {
	# Report the error
	showText $result
	return 1
    } else {
	# Display the results in user's EDITOR
	set fname [file join /tmp [pid]_rev${r}_$key]
	close [open $fname w]
	file attributes $fname -permissions u=rw,go=
	set fid [open $fname w]
	puts $fid $bg_output
	close $fid
	system "($editor $fname;rm $fname)" &
    }
    catch "exec rlog -zLT -r$r $key" rlogResult
    showText "$rlogResult"
    return 0
}

# Check in and specify revision.
itcl::body RCS::RCScheckinNew {} {
    showText "Checking in with with user-specified revision..."
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }

    # Ask user which selection they want
    if {[getRevision 1]} {
	showText "Check in canceled."
	return 1
    }
    set r $revSelect

    # Try and get a log message
    if {[getDescription "Enter log message"]} {
	# Canceled, exit
	showText "Check in canceled."
	return 1
    } else {
	set logText $descText
    }

    # Check in the revision.
    update idletasks
    catch "exec ci -zLT $rev \"-m$logText\" -f -u$r $key < /dev/null" result
    showText "$result"
    getRCSinfo
    return 0
}

# Delete (outdate) older revisions.
itcl::body RCS::RCSoutDate {} {
    showText "Deleting (outdating) old revisions..."
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }

    # Ask user which selection they want
    #   (allow multiple, but continuous selections)
    if {[getRevision 0 multiple]} {
	showText "Outdating canceled."
	return 1
    }
    set r "[lindex $revSelect end]:[lindex $revSelect 0]"

    # Check out old revision and run xless
    update idletasks
    catch "exec rcs -o$r $key < /dev/null" result
    showText "$result"
    getRCSinfo
    return 0
}

# Set the default branch.
itcl::body RCS::RCSsetBranch {} {
    showText "Setting default branch..."
    # Get RCS file
    set key [getCurrent]
    if {![string length $key]} {
	return 1
    }

    # Ask user which selection they want
    if {[getRevision]} {
	showText "Set default branch cancelled."
	return 1
    }
    set r $revSelect

    # Check out old revision and run xless
    catch "exec rcs -b$r $key" rcsResult
    showText "$rcsResult"
    getRCSinfo
    return 0
}

########################################################################
#                                                                      #
# DIALOGS                                                              #
#                                                                      #
########################################################################

# Check in a new file.
itcl::body RCS::RCSnew {} {
    #
    # Check for RCS directory, allow user to create
    #
    if {![file exists $dir/RCS]} {
	# Wait cursor
	animCursor::resetCursor
	animCursor::animCursor $top
	# RCS does not exist, so user must either create or exit
	set createRCS [MessageDlg $top.create \
		-title "Create RCS?" \
		-message "Should I create $dir/RCS?" \
		-icon question \
		-type yesno]
	# Clear cursor, global grab
	animCursor::resetCursor
	grab .
	if {!$createRCS} {
	    # User has asked to create, so we try that
	    if {[catch "file mkdir $dir/RCS" result]} {
		# We FAILED to make the directory, so we return
		catch "puts stderr {$result}"
		set statusMsg "Failed to mkdir $dir/RCS"
		$status configure -foreground red
		return 1
	    } else {
		# Directory was created, so we set permissions
		file attributes $dir/RCS -permissions 040775
	    }
	} else {
	    # User does not want to create RCS directory, so we return
	    set statusMsg "$dir/RCS does not exist!"
	    $status configure -foreground red
	    return 1
	}
    }

    #
    # Find files not already in RCS directory
    #

    catch "unset noRCS"
    # Find ALL files and noRCS array to 1
    set AllFiles [lsort [glob -nocomplain -- $dir/*]]
    foreach f $AllFiles {
	set base [lindex [split $f "/"] end]
	set noRCS($base) 1
    }
    # Find RCS-files and set noRCS array to 0
    set RCSfiles [lsort \
	[glob -nocomplain -- $dir/RCS/* $dir/RCS/.\[a-zA-Z0-9\]*]]
    foreach f $RCSfiles {
	regsub -- {,v$} [lindex [split $f "/"] end] "" base
	set noRCS($base) 0
    }

    # Allow user to pick from files where noRCS array = 1
    destroy "$top.create"
    toplevel $top.create -relief ridge -class dialog
    wm title $top.create "Select file"
    tixScrolledHList $top.create.listbox \
	-options "
	    hlist.command	{[itcl::code $this RCScreate]}
	    hlist.columns	1
	    hlist.header	1
	    hlist.selectMode	extended
	    hlist.width 	40
	    hlist.height	20
	    hlist.separator	/
	"
    $top.create.listbox configure -highlightcolor blue
    set nhlist [$top.create.listbox subwidget hlist]
    $top.create.listbox.f1 configure -borderwidth 2 -relief sunken

    # Wheel mouse
    bind $nhlist <Button-4> "%W yview scroll -5 units ; break"
    bind $nhlist <Button-5> "%W yview scroll 5 units ; break"
    bind $nhlist <Shift-Button-4> "%W yview scroll -1 units ; break"
    bind $nhlist <Shift-Button-5> "%W yview scroll 1 units ; break"
    bind $nhlist <Control-Button-4> "%W yview scroll -1 pages ; break"
    bind $nhlist <Control-Button-5> "%W yview scroll 1 pages ; break"
    bind $nhlist <Alt-Button-4> "%W xview scroll -5 units ; break"
    bind $nhlist <Alt-Button-5> "%W xview scroll 5 units ; break"

    # Configure Headings
    set h_style [tixDisplayStyle text -refwindow $nhlist -fg black]
    $nhlist header create 0 -itemtype text -text "Directory:\n$dir" \
	-style $h_style
    $nhlist header configure 0 -relief groove
    $nhlist column width 0 -char 40

    # Load listbox
    foreach key [lsort [array names noRCS]] {
	if {$noRCS($key)} {
	    # Exclude directories
	    if {![file isdirectory $dir/$key]} {
		$nhlist add $key -itemtype text -text $key
	    }
	}
    }

    # Buttons
    set b [frame $top.create.buttons -relief sunken -borderwidth 2]
    button $b.ok -text "OK" -takefocus 1 \
	-highlightthickness 2 -highlightcolor blue \
	-command "
	    animCursor::resetCursor
	    [itcl::code $this RCScreate] \[$nhlist select get\]
	"
    button $b.cancel -text "Cancel" -takefocus 1 \
	-highlightthickness 2 -highlightcolor blue \
	-command "
	    animCursor::resetCursor
	    destroy $top.create
	"
    catch "imageLib::imageConfig $b.ok OK -relief flat"
    catch "imageLib::imageConfig $b.cancel Cancel -relief flat"

    # Pack
    pack $b -side bottom -fill x -expand 0
    pack $b.ok -side left -padx 20 -pady 10
    pack $b.cancel -side right -padx 20 -pady 10
    pack $top.create.listbox -side top -fill y -expand 1 -anchor w

    # Wait cursor, grab and return
    animCursor::resetCursor
    animCursor::animCursor $top
    grab $top.create
    return 0
}

# This proc does the actual initial check in
itcl::body RCS::RCScreate {args} {
    # Destroy the dialog
    destroy $top.create

    # Check to see if any keys were passed and if directory exists
    if {![string length $args]} {
	animCursor::resetCursor
	return 1
    }
    if {[catch "cd $dir"]} {
	catch "puts stderr {$dir does not exist}"
	animCursor::resetCursor
	return 1
    }
    eval set allKeys $args

    # Loop through keys and attempt to add
    foreach key $allKeys {
	# Does file exist?
	if {![file exists $dir/$key]} {
	    catch "puts stderr {$dir/$key does not exist}"
	    animCursor::resetCursor
	    return 1
	}

	# Is it a file?
	if {![file isfile $dir/$key]} {
	    catch "puts stderr {$dir/$key is not a regular file}"
	    animCursor::resetCursor
	    return 1
	}

	# Show file and check in...
	system "xless $dir/$key" &
	showText "Checking $key in..."
	update idletasks
	system "xterm -geometry +0+0 -e ci -zLT $rev -u $key"

	# Update RCSinfo
	getRCSinfo
    }
    animCursor::resetCursor
    return 0
}

# Prompt user and change current directory.
itcl::body RCS::changeDir {} {
    # Try and move into current directoy
    catch "cd $dir"
    # Build Directory Select dialog, if not already done
    if {![winfo exists $top.cd]} {
	#
	# Build widgets
	#
	toplevel $top.cd -class Dialog -relief ridge
	wm withdraw $top.cd
	wm title $top.cd "Select Directory"
	# Combo box for entering directory or choosing from history
	tixComboBox $top.cd.cbox -dropdown true -editable true \
	    -labelside top -label "Directory:" \
	    -variable "[itcl::scope _tmpDir]" \
	    -command "[itcl::code $this setCBox]" \
	    -options {
		listbox.height 10
		label.width 10
		label.anchor w
	    }
	[$top.cd.cbox subwidget entry] configure -takefocus 1 \
	    -borderwidth 2 -highlightthickness 2 -highlightcolor blue
	[$top.cd.cbox subwidget listbox] configure -background white
	# Directory list
	tixDirList $top.cd.dir -value [exec pwd] \
	    -browsecmd "set {[itcl::scope _tmpDir]}" \
	    -command "set {[itcl::scope _tmpDir]}"
	[$top.cd.dir subwidget hlist] configure \
	    -selectmode browse
	$top.cd.dir.f1 configure -highlightcolor blue
	# Buttons
	set b [frame $top.cd.buttons -relief sunken -borderwidth 2]
	button $b.ok -text OK -relief raised -takefocus 1 \
	    -borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	    -command "[itcl::code $this dirOK]"
	button $b.cancel -text Exit -relief raised -takefocus 1 \
	    -borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	    -command "
		[itcl::code $this getRCSinfo]
		animCursor::resetCursor
		grab .
		wm withdraw $top.cd
	    "
	# Pack widgets
	pack $b -side bottom -fill x -expand 0
	pack $b.ok -side left -padx 20 -pady 10
	pack $b.cancel -side right -padx 20 -pady 10
	pack $top.cd.cbox -side top -fill both -expand 0
	pack $top.cd.dir -side top -fill both -expand 1

	# Use images from imageLib, if available
	catch "imageLib::imageConfig $top.cd.buttons.ok OK \
	    -relief flat"
	catch "imageLib::imageConfig $top.cd.buttons.cancel Cancel \
	    -relief flat"
    }

    # Load the history of the combo box
    [$top.cd.cbox subwidget listbox] delete 0 end
    if {[catch {read_file -nonewline ~/.tkrcs-dirs} dirHist]} {
	set dirHist {}
    }
    set dirHist [lrmdups [split "$dir\n$dirHist" "\n"]]
    foreach h $dirHist {
	if {[string length $h]} {
	    [$top.cd.cbox subwidget listbox] insert end "$h"
	}
    }

    # Show dialog, set wait cursor and grab
    catch "$top.cd.dir configure -value $dir"
    wm deiconify $top.cd
    animCursor::resetCursor
    animCursor::animCursor $top
    grab $top.cd
    update idletasks
    wm geometry $top.cd [wm geometry $top.cd]
    return 0
}

# Prompt user and change symbolic name.
itcl::body RCS::changeRev {} {
    # Prompt user for new symbolic name
    set input .[namespace tail $this]_input
    destroy $input
    set newRev [[GetInput $input \
	    -prompt "New revision name:" \
	    -value [string range [cget -rev] 2 end]] activate]

    # Reset symbolic name if not cancelled or null
    if {[string length $newRev]} {
	configure -rev [lindex $newRev 0]
    }
}

# Validate change to the combo box.
itcl::body RCS::setCBox {{d {}}} {
    catch "$top.cd.dir configure -value $d"
    set _tmpDir [$top.cd.dir cget -value]
    return
}

# Called by the OK button in changeDir.
itcl::body RCS::dirOK {} {
    # Insert date into dir widget (which will validate it)
    # then use the date retrieved...
    setCBox [[$top.cd.cbox subwidget entry] get]
    if {[catch {configure -dir [$top.cd.dir cget -value]}]} {
	return 1
    }

    # Add the dir to the history and write it out
    [$top.cd.cbox subwidget listbox] insert end $dir
    set dirHist [join [lrmdups \
	[[$top.cd.cbox subwidget listbox] get 0 end]] "\n"]
    set histFile [open ~/.tkrcs-dirs w]
    puts $histFile "$dirHist"
    close $histFile

    # Normal cursor, ungrab and withdraw
    animCursor::resetCursor
    grab .
    wm withdraw $top.cd
    return 0
}

# Allow user to enter or edit a description (change log).
itcl::body RCS::getDescription {prompt {curText {}}} {
    # Toplevel widget
    destroy $top.desc
    toplevel $top.desc -class Dialog -relief groove
    wm withdraw $top.desc
    wm title $top.desc "$prompt"

    # Prompt
    label $top.desc.prompt -text "$prompt" -anchor w -font 12x24

    # Text
    tixScrolledText $top.desc.text
    set t [$top.desc.text subwidget text]
    $t configure -wrap none -takefocus 1 -width 75 -height 10 \
	-highlightcolor blue -highlightthickness 2
    $t insert insert $curText

    # Buttons
    set b [frame $top.desc.buttons -relief sunken -borderwidth 2]
    button $b.ok -text OK -relief raised -takefocus 1 \
	-borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	-command "
	    set {[itcl::scope descReturn]} 0
	    set {[itcl::scope descText]} \[[itcl::code $this getDescText] $t\]
	    destroy $top.desc
	"
    button $b.cancel -text Cancel -relief raised -takefocus 1 \
	-borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	-command "
	    set {[itcl::scope descReturn]} 1
	    set {[itcl::scope descText]} \[[itcl::code $this getDescText] $t\]
	    destroy $top.desc
	"
    catch "imageLib::imageConfig $b.ok OK -relief flat"
    catch "imageLib::imageConfig $b.cancel Cancel -relief flat"

    # Pack widget
    pack $b -side bottom -fill x -expand 0
    pack $b.ok -side left -padx 20 -pady 10
    pack $b.cancel -side right -padx 20 -pady 10
    pack $top.desc.prompt -side top -fill both -expand 0
    pack $top.desc.text -side top -fill both -expand 1

    # Wait cursor, grab, wait, then return
    wm deiconify $top.desc
    wm geometry $top.desc +0+0
    update idletasks
    set descReturn 1
    animCursor::resetCursor
    animCursor::animCursor $top
    grab $top.desc
    focus $t
    tkwait window $top.desc
    animCursor::resetCursor
    return $descReturn
}

# Filter text from the widget and make required quoting # substitions.
itcl::body RCS::getDescText {t} {
    # Get the text
    set descText [$t get 0.0 end]
    # Escape special characters and remove trailing newlines
    regsub -all "\"" $descText "\\\"" descText
    regsub -all {\$} $descText {\\$} descText
    regsub "\n+$" $descText {} descText
    # If text is empty, give it a default message
    if {![string length $descText]} {
	set descText "no message"
    }

    return $descText
}

# Allow user to select a revision or enter a new one.
itcl::body RCS::getRevision {{newRev 0} {browseMode browse}} {
    # Set key (return if none)
    set key [$hlist selection get]
    if {![string length $key]} {
	return 1
    }

    # Toplevel widget
    destroy $top.rev
    toplevel $top.rev -class Dialog -relief groove
    wm withdraw $top.rev
    wm title $top.rev "Select Revision"

    # Entry
    if {$newRev} {
	set Prompt "Enter new revision:"
    } else {
	set Prompt "Select revision:"
    }
    tixLabelEntry $top.rev.entry -label $Prompt -labelside left
    set e [$top.rev.entry subwidget entry]
    $e configure -takefocus 1 -highlightcolor blue -highlightthickness 2

    # Historical Revisions
    tixScrolledHList $top.rev.history \
	-options "
	    hlist.command	{}
	    hlist.columns	2
	    hlist.header	1
	    hlist.selectMode	$browseMode
	    hlist.width 	40
	    hlist.height	20
	    hlist.separator	/
	"
    $top.rev.history configure -highlightcolor blue
    set rhlist [$top.rev.history subwidget hlist]
    $rhlist configure -browsecmd "
	$e delete 0 end
	$e insert 0 \[$rhlist selection get\]
	# Ignore key passed here..."
    $top.rev.history.f1 configure -borderwidth 2 -relief sunken

    # Wheel mouse
    bind $rhlist <Button-4> "%W yview scroll -5 units ; break"
    bind $rhlist <Button-5> "%W yview scroll 5 units ; break"
    bind $rhlist <Shift-Button-4> "%W yview scroll -1 units ; break"
    bind $rhlist <Shift-Button-5> "%W yview scroll 1 units ; break"
    bind $rhlist <Control-Button-4> "%W yview scroll -1 pages ; break"
    bind $rhlist <Control-Button-5> "%W yview scroll 1 pages ; break"
    bind $rhlist <Alt-Button-4> "%W xview scroll -5 units ; break"
    bind $rhlist <Alt-Button-5> "%W xview scroll 5 units ; break"

    # Configure headings
    set h_style [tixDisplayStyle text -refwindow $rhlist \
	-fg black -anchor center]
    $rhlist header create 0 -itemtype text -text "Rev." -style $h_style
    $rhlist header create 1 -itemtype text -text "Details" -style $h_style
    $rhlist header configure 0 -relief groove
    $rhlist header configure 1 -relief groove
    $rhlist column width 0 -char 7
    $rhlist column width 1 -char 33

    # Load listbox
    foreach r $_rlog($key,revHist) {
	if {$newRev} {
	    set State "disabled"
	} else {
	    set State "normal"
	}
	$rhlist add [lindex $r 0] \
	    -itemtype text \
	    -text "[lindex $r 0]\n" \
	    -state $State
	set desc [join [split [lindex [split $_rlog($key,$r) \n] 0] {;}] \n]
	if {[llength [split $_rlog($key,$r) \n]] > 1} {
	    append desc "[lindex [split $_rlog($key,$r) \n] 1]\n"
	}
	append desc "    [string repeat - 20]"
	$rhlist item create [lindex $r 0] 1 \
	    -itemtype text \
	    -text "  $desc"
    }

    # Buttons
    set b [frame $top.rev.buttons -relief sunken -borderwidth 2]
    button $b.ok -text OK -relief raised -takefocus 1 \
	-borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	-command "
	    set {[itcl::scope revReturn]} 0
	    set {[itcl::scope revSelect]} \[$e get\]
	    destroy $top.rev
	"
    button $b.cancel -text Cancel -relief raised -takefocus 1 \
	-borderwidth 2 -highlightthickness 2 -highlightcolor blue \
	-command "
	    set {[itcl::scope revReturn]} 1
	    set {[itcl::scope revSelect]} \[$e get\]
	    destroy $top.rev
	"
    catch "imageLib::imageConfig $b.ok OK -relief flat"
    catch "imageLib::imageConfig $b.cancel Cancel -relief flat"

    # Bindings
    bind $top.rev <Return> "
	set {[itcl::scope revReturn]} 0
	set {[itcl::scope revSelect]} \[$e get\]
	destroy $top.rev
    "

    # Pack widget
    pack $b -side bottom -fill x -expand 0
    pack $b.ok -side left -padx 20 -pady 10
    pack $b.cancel -side right -padx 20 -pady 10
    pack $top.rev.entry -side top -fill both -expand 0
    pack $top.rev.history -side top -fill both -expand 1

    # Wait cursor, grab, wait, then return
    wm deiconify $top.rev
    wm geometry $top.rev +0+0
    update idletasks
    set revReturn 1
    animCursor::resetCursor
    animCursor::animCursor $top
    grab $top.rev
    tkwait window $top.rev
    animCursor::resetCursor
    return $revReturn
}

# Allow a user to select a file in the current directory.
itcl::body RCS::selectFile {fileName} {
    $hlist selection set $fileName
    return
}
